#!/usr/bin/python2
"""This script modifies the PNP XML files to refer to "multiple" RRD files. It
also executes `rrdsplit` on your "single" RRD files to split the storage into
one rrd file per datasource."""

# python 2.6 compatible.. :[ argparse is not available
import optparse
import os
import re
import subprocess
import sys
import time
from xml.etree import ElementTree as ET

# Enforce UTF-8 as encoding of printed messages, regardless of tty presence
# and/or terminal locale settings.
import codecs
sys.stdout = codecs.getwriter('UTF-8')(sys.stdout)

# Inspired by http://stackoverflow.com/a/4814985
def proc_run_catch_stdouterr(*popenargs, **kwargs):
    if 'stdout' in kwargs:  # pragma: no cover
        raise ValueError('stdout argument not allowed, '
                         'it will be overridden.')
    process = subprocess.Popen(stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                               *popenargs, **kwargs)
    output, _ = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise ProcRunCatchError(retcode, cmd,
                                output=output)
    return output

class ProcRunCatchError(Exception):
    def __init__(self, returncode, cmd, output=None):
        self.returncode = returncode
        self.cmd = cmd
        self.output = output

    def __str__(self):
        return "Command '%s' returned non-zero exit status %d" % (
            self.cmd, self.returncode)


class PNPXMLConversionException(Exception): pass
class PNPXMLConversionIgnored(Exception): pass


def msg(msg, log=False):
    # Once printed, the unicode string (u'...'), will be encoded into UTF-8,
    # due to our default encoding set above. If our msg string isn't unicode,
    # just an ordinary string, we first need to decode the UTF-8 data into
    # unicode. If we encounted non-UTF-8-data while doing this, such as extended
    # ascii, this will be replaced by the "replacement char".
    if(isinstance(msg, str)):
        msg = msg.decode("utf-8", "replace")

    if log:
        print("%s [%s] %s" % (time.strftime('%Y-%m-%d %H:%M:%S %z'), os.getpid(), msg))
    else:
        print(msg)

def convert_pnp_xml(pnp_xml_path, opts):
    msg("processing: %s" % pnp_xml_path, opts.log)
    if not os.path.exists(pnp_xml_path):
        raise PNPXMLConversionException("%s does not exist" % pnp_xml_path)
    pnp_stat = os.stat(pnp_xml_path)

    try:
        pnp_xml = ET.parse(pnp_xml_path)
    except Exception as ex:
        raise PNPXMLConversionException("Failed to parse PNP XML (%s): %s" % (pnp_xml_path, str(ex)))

    # Encode as utf8 encoded strings, as this is how the files are stored
    # in the filesystem.
    rrdfiles = [rrdfile.text.encode('utf-8') for rrdfile in pnp_xml.findall('./DATASOURCE/RRDFILE')]
    if len(rrdfiles) == 0:
        raise PNPXMLConversionIgnored('No need to convert %s, since it has no RRDFILE specified' % (pnp_xml_path))

    rrd_out_files = []
    for ds in pnp_xml.findall('DATASOURCE'):
        # Same as above - utf8 encoded strings.
        ds_name = ds.find('NAME').text.strip().encode('utf-8')
        clean_ds_name = re.sub(r'[ :\/]', r'_', ds_name)
        rrd_storage_type = ds.find('RRD_STORAGE_TYPE')
        if rrd_storage_type is None:
            raise PNPXMLConversionException("No RRD_STORAGE_TYPE specified for '%s' in %s, cannot convert!" % (ds_name, pnp_xml_path))

        if rrd_storage_type.text.strip() == 'MULTIPLE':
            raise PNPXMLConversionIgnored("No need to convert DATASOURCE '%s' in %s, since its RRD_STORAGE_TYPE is already MULTIPLE" % (ds_name, pnp_xml_path))
        ds_idx = int(ds.find('DS').text.strip())

        # fill up with undefined DS outputs up until this index
        while len(rrd_out_files) < ds_idx:
            rrd_out_files.append("")

        # use all of the original, up until the extension and append '_<name>'.rrd to that
        rrd_out_files[ds_idx - 1] = rrdfiles[0].rsplit('.rrd', 1)[0] +  '_' + clean_ds_name + '.rrd'

    if not all(rrdfiles[0] == rrdfile for rrdfile in rrdfiles):
        raise PNPXMLConversionException("found mismatching RRDFILE declaration in %s" % pnp_xml_path)

    if not os.path.exists(rrdfiles[0]):
        raise PNPXMLConversionException("RRD file %s, declared in %s, does not exist on this system" % (rrdfiles[0], pnp_xml_path))

    rrdsplit_cmd = ['rrdsplit']
    if opts.purge:
        rrdsplit_cmd.append('-d')

    rrdsplit_cmd.append('-o')
    rrdsplit_cmd.append(','.join(rrd_out_files))

    rrdsplit_cmd.append(rrdfiles[0])
    msg("running: " + ' '.join(rrdsplit_cmd).strip(), opts.log)

    try:
        output = proc_run_catch_stdouterr(rrdsplit_cmd)
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            raise PNPXMLConversionException("rrdsplit does not exist on PATH")
    except ProcRunCatchError as e:
        raise PNPXMLConversionException("Failed rrdsplit execution (code: %s) (output: %s)" % (e.returncode, e.output))

    msg("success:\n%s" % output, opts.log)

    for rrd_out, ds in zip(rrd_out_files, pnp_xml.findall('DATASOURCE')):
        try:
            os.chown(rrd_out, pnp_stat.st_uid, pnp_stat.st_gid)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                msg("info: RRD %s does not exist after split, likely stale data in source" % rrd_out, opts.log)
                continue
        # As the xml .write() wants unicode to properly write as utf8 in the
        # file, let's give it unicode now, instead of utf8, so .write() can
        # make it utf8 on its own. >_>
        ds.find('RRDFILE').text = rrd_out.decode('utf-8')
        ds.find('RRD_STORAGE_TYPE').text = 'MULTIPLE'
        ds.find('DS').text = '1'

    with open(pnp_xml_path, 'w') as f:
        pnp_xml.write(f, encoding='UTF-8')

def main():
    parser = optparse.OptionParser(prog="mon rrd-multi-storage", description=__doc__)
    parser.add_option('--purge', action='store_true', dest='purge', default=False,
            help='Purge the old, non-transformed and non-usable rrd-files.' +
            ' Could be used for journaling')
    parser.add_option('--log', action='store_true', dest='log', default=False,
            help='Messages are printed like a log file (time + pid).')

    opts, args = parser.parse_args()

    try:
        if len(args) > 0:
            convert_pnp_xml(args[0], opts)
    except PNPXMLConversionException as e:
        msg("error: %s" % str(e), opts.log)
        sys.exit(1)
    except PNPXMLConversionIgnored as e:
        msg("ignored: %s" % str(e), opts.log)
        sys.exit(0)

if __name__ == '__main__':
    main()
