#!/bin/bash

IFS=$' \t\n'
rrdcached_pidfile=/opt/monitor/var/rrdtool/rrdcached/rrdcached.pid
perfdata_path=/opt/monitor/op5/pnp/perfdata
perfdata_spool_path=/opt/monitor/var/spool/perfdata
process_perfdata_cfg_path=/opt/monitor/etc/pnp/process_perfdata.cfg
process_perfdata_path=/opt/monitor/op5/pnp/process_perfdata.pl
ts=$(date +%s)
runtime_dir=$perfdata_path/.rrd-storage-convert
backup_path=$runtime_dir/$ts.backup.tar.gz
convert_log=$runtime_dir/$ts.log
cancelled_path=$runtime_dir/$ts.converted
PROG=$(basename $0)

function syntax() {
  cat << 'EOF'
Usage: mon rrd-storage-convert

This command is a wrapper around `mon sysconf rrd-multi-storage' which tries
to be as careful and user-friendly as possible while converting all PNP
service configuration XML's on the system. If you want to convert a system
to use MULTIPLE RRD_STORAGE_MODE, this is probably what you want.
EOF
  exit ${1:-1}
}

function flush_and_restart_rrdcached() {
  rrdcached_pid=$(cat ${rrdcached_pidfile} 2>/dev/null)
  [ "$rrdcached_pid" -gt '1' ] &> /dev/null || \
    err_exit "Unable to get rrdcached pid. Is rrdcached not running?"

  progress "Flushing and stopping rrdcached ..."
  kill -USR1 ${rrdcached_pid} || err_exit "Failed to flush rrdcached!"
  progress "Waiting for rrdcached to stop ..."

  while kill -0 ${rrdcached_pid} 2>/dev/null; do
    sleep 1s
    echo -n '.'
  done
  echo ''

  progress "Starting rrdcached post-flush ..."
  service rrdcached start || err_exit "Failed to start rrdcached!"
  echo ''
}

function progress() {
  echo "${1:-"..."}"
}

function err_exit() {
  echo "ERROR: ${1:-"Unknown error!"}" 1>&2
  exit 1
}

function ask() {
  while true; do
    read -p "${1:-'Huh?'} (Yes/No) " yesorno
    case $yesorno in
      [Yy]* ) return 0;;
      [Nn]* ) return 1;;
      * ) echo "Please answer yes or no.";;
    esac
  done
}

function find_perf()
{
  find . -mindepth 1 -maxdepth 1 \
    ! -path './.rrd-storage-convert' \
    ! -path './.trash' \
    -print0
}

function get_spool_files()
{
  ls -1 ${perfdata_spool_path} | grep -v -E -- '-PID-[[:digit:]]+$'
}

function print_stats()
{
  log_found="$(grep -c '\] processing:' "$convert_log")"
  log_error="$(grep -c '\] error:' "$convert_log")"
  log_success="$(grep -c '\] success:' "$convert_log")"
  log_ignored="$(grep -c '\] ignored:' "$convert_log")"
  log_info="$(grep -c '\] info:' "$convert_log")"

  printf '%10s XML file(s) found.\n' "$log_found"
  printf "%10s (keyword: processing)\n\n" ''

  printf '%10s successfully processed XML file(s) (related RRD files converted).\n' "$log_success"
  printf "%10s (keyword: success)\n\n" ''

  if [ "$log_error" -gt '0' ]; then
    printf '%10s faulty XML file(s) found, but this should be harmless as the files\n' "$log_error"
    printf '%10s were likely already invalid prior to running this tool.\n' ''
    printf "%10s (keyword: error)\n\n" ''
  fi

  if [ "$log_ignored" -gt '0' ]; then
    printf '%10s ignored XML file(s) since no related RRD files were found, or the\n' "$log_ignored"
    printf '%10s related RRD files had already been converted.\n' ''
    printf "%10s (keyword: ignored)\n\n" ''
  fi

  if [ "$log_info" -gt '0' ]; then
    printf '%10s RRD file(s) not found after being converted. This is probably\n' "$log_info"
    printf '%10s harmless, and due to old, stale and no longer used data.\n' ''
    printf "%10s (keyword: info)\n\n" ''
  fi

  printf 'More detailed log data can be found in:\n%s\n\n' "$convert_log"
}



if [ "$1" == '--help' -o "$1" == '-h' ]; then
  syntax 0
fi

if [ $EUID -ne 0 ]; then
  err_exit "$PROG should be run as root"
fi

grep -E -q -e '^[[:space:]]*RRD_STORAGE_TYPE[[:space:]]*=.*' ${process_perfdata_cfg_path} || err_exit "No RRD_STORAGE_TYPE configured in ${process_perfdata_cfg_path}, cannot continue"

grep -E -e '^[[:space:]]*RRD_STORAGE_TYPE[[:space:]]*=[[:space:]]*MULTIPLE' ${process_perfdata_cfg_path}
if [ $? -eq 0 ]; then
  if ! ask "It looks like this system is already configured to use RRD_STORAGE_TYPE multiple. Would you like to proceed anyway?"; then
    err_exit "Aborting."
  fi
fi

progress "Stopping NPCD ..."
if service npcd status > /dev/null; then
  service npcd stop || err_exit "Failed to stop NPCD!"
else
  progress "NPCD already stopped."
fi

flush_and_restart_rrdcached

mkdir -p "$runtime_dir" || err_exit "Failed creating runtime directory ($runtime_dir)."

progress "Calculating required and available disk space for backup ..."
required_disk_space_approx=$(du -sk ${perfdata_path} | awk '{print $1*2}')
available_disk_space=$(df -P -k ${perfdata_path} | awk '/[0-9]%/{print $(NF-2)}')

if [ ${available_disk_space} -lt ${required_disk_space_approx} ]; then
  if ! ask "There might not be enough disk space available to create perfdata backup. Need approximately ${required_disk_space_approx} kB, but only ${available_disk_space} kB available. Would you like to proceed anyway?"; then
    err_exit "Aborting."
  fi
fi

progress "Backing up perfdata to ${backup_path} ..."
# Subshell for temporary change of cwd.
(
  cd ${perfdata_path} || err_exit "Failed entering perfdata directory '$perfdata_path'."
  # 1) Print all items found in cwd except the ignored paths, null separated.
  # 2) Sort items, still null separated.
  # 3) Create a tarball based on the null separated file list on stdin.
  find_perf | sort -z | \
    tar -c -z -f ${backup_path} --checkpoint=1000 --checkpoint-action=dot \
    --null --files-from - || err_exit "Failed to create backup."
) || exit $?

echo ''
echo ''
progress "Converting RRDs (this could take a while) ..."
# For each xml file in $perfdata_path, run "mon sysconf rrdmultistorage" on it
# (in up to $(nproc) processes in parallel), put the output in rrdmultistorage,
# and output to STDOUT all "processing" lines to communicate progress
# A depth of 2 will make sure we ignore the .trash dir and other unknown stuff.
find ${perfdata_path} -mindepth 2 -maxdepth 2 -type f -name '*.xml' -print0 | \
  xargs -0 -n 1 -P $(nproc) mon sysconf rrd-multi-storage --purge --log 2>&1 | \
  tee -a "$convert_log" | \
  grep '\] processing:'

progress "Done ..."

print_stats
cat << EOF

** ATTENTION **

Conversion is not yet finished. Please browse some of the graphs in your
op5 Monitor GUI, and make sure that you are not only seeing red error
messages instead of graphs.

If the graphs are not working as expected, you may restore the original graphs.

NOTE: It is expected that graph data generated *during* the runtime of this
tool will *not* be seen in your graphs. It will be seen later on.

EOF
if ask 'Continue and complete RRD conversion (yes), or restore original graphs from backup (no)?'; then
  echo ''
  progress "Updating global RRD_STORAGE_TYPE in ${process_perfdata_cfg_path}..."
  sed -i -r -e 's/^[[:space:]]*(RRD_STORAGE_TYPE)[[:space:]]=.*$/\1 = MULTIPLE/' ${process_perfdata_cfg_path} || err_exit "Failed to update global RRD_STORAGE_TYPE setting in ${process_perfdata_cfg_path}!"
  restored=''

else
  mkdir -p "$cancelled_path" || err_exit "Failed creating directory: $cancelled_path"
  # Subshell for temporary change of cwd.
  (
    echo ''
    progress "Entering $perfdata_path ..."
    cd "$perfdata_path" || err_exit "Failed entering perfdata directory '$perfdata_path'."

    # Move all dirs that may contain the converted files into the
    # "cancelled dir", allowing one to analyze what went wrong afterwards.
    progress "Storing converted data in $cancelled_path ..."
    find_perf | xargs -0 -I% -n1 mv '%' "$cancelled_path/"

    # Print a dot for every dir found and count the number of dots (=none!).
    files="$(find_perf | xargs -0 -I% -n1 printf . | wc -c)"
    [ "$files" == '0' ] || \
      err_exit "Failed moving $files item(s) into '$cancelled_path/'. Contact op5 Support"

    progress "Restoring backup file $backup_path ..."
    tar -x -z -f "$backup_path" --checkpoint=1000 --checkpoint-action=dot || \
      err_exit "Failed restoring backup."
  ) || exit $?
  cat << EOF


Successfully restored backup. Please contact op5 Support if you want to analyze
what went wrong while converting your RRD storage. Debug data is found in:
$runtime_dir
EOF
  restored='yes'
fi

echo ''
while :; do
  backlog_count="$(get_spool_files | wc -l)"
  [ ${backlog_count} -gt 20 ] || break
  progress "${backlog_count} files accumulated in perfdata backlog, processing to avoid congestion ..."

  # stderr to devnull due to annoying write errors when we pipe into head and sudo
  get_spool_files 2>/dev/null | head -n1000 | \
    asmonitor xargs -I% -n 1 ${process_perfdata_path} -t 10000 -b "${perfdata_spool_path}/%"
done

echo ''
progress "Starting NPCD ..."
service npcd start || err_exit "Failed to start NPCD!"
if [ "$restored" != 'yes' ]; then
  cat << EOF
Conversion from single to multiple RRD storage done!

If everything works as expected, you might want to remove backup data found in:
$runtime_dir
EOF
fi
